"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const grpc_1 = __importDefault(require("grpc"));
const glob_1 = __importDefault(require("glob"));
const callStub_1 = require("./callStub");
const logger_1 = __importStar(require("./logger"));
const proto_loader_1 = require("@grpc/proto-loader");
class GrpcAsyncServer {
    constructor(config) {
        this.config = config;
        if (config.logging) {
            logger_1.setLogger(config.logging);
        }
        this.server = new grpc_1.default.Server(config.grpcOptions);
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            this.addServices();
            const { binding, port } = this.config;
            const address = `${binding || '127.0.0.1'}:${port}`;
            this.server.bind(address, this.getCredentials());
            this.server.start();
            logger_1.default.info(`Listening on ${address}`);
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise(done => this.server.tryShutdown(done));
        });
    }
    getCredentials() {
        const { ssl } = this.config;
        if (ssl) {
            return grpc_1.default.ServerCredentials.createSsl(ssl.rootCerts, ssl.keyCertPairs, ssl.checkClientCertificate);
        }
        else {
            return grpc_1.default.ServerCredentials.createInsecure();
        }
    }
    getProtoFiles() {
        const { protos, options: { includeDirs }, } = this.config;
        return (includeDirs || [process.cwd()])
            .map((root) => {
            return protos
                .map(pattern => glob_1.default.sync(pattern, { root, cwd: root, silent: true, absolute: true }))
                .reduce((a, b) => [...a, ...b], []);
        })
            .reduce((a, b) => [...a, ...b], [])
            .filter(name => name.endsWith('.proto'));
    }
    getServicesFromProto() {
        const allFiles = this.getProtoFiles();
        const serviceDefinitions = {};
        for (const fileName of allFiles) {
            logger_1.default.debug(`Loading proto: ${fileName.replace(process.cwd(), '')}`);
            const protoPackage = proto_loader_1.loadSync(fileName, this.config.options);
            for (const serviceName in protoPackage) {
                const serviceDef = protoPackage[serviceName];
                const isServiceDef = Object.keys(serviceDef)
                    .map(member => serviceDef[member])
                    .filter(o => (o.path && o.requestType) || o.responseType).length >= 1;
                if (isServiceDef) {
                    serviceDefinitions[serviceName] = serviceDef;
                }
            }
        }
        return serviceDefinitions;
    }
    registerService(serviceName, serviceDef, serviceDefinition) {
        const serviceImpl = typeof serviceDefinition === 'object' ? serviceDefinition : new serviceDefinition();
        const members = Object.keys(serviceDef).filter(name => typeof serviceDef[name].path === 'string');
        logger_1.default.debug(`Loading service: ${serviceName} with ${serviceImpl.constructor.name}`);
        const svcFunctionMap = members.reduce((obj, method) => {
            obj[method] = callStub_1.callStub(this.server, serviceName, serviceImpl, method);
            return obj;
        }, {});
        this.server.addService(serviceDef, svcFunctionMap);
    }
    addServices() {
        const serviceDefinitions = this.getServicesFromProto();
        for (const serviceName in this.config.services) {
            const serviceDef = serviceDefinitions[serviceName];
            if (!serviceDef) {
                logger_1.default.error(`Service not found: "${serviceName}".`);
                process.exit(1);
            }
            delete serviceDefinitions[serviceName];
            this.registerService(serviceName, serviceDef, this.config.services[serviceName]);
        }
        const missing = Object.keys(serviceDefinitions);
        if (missing.length) {
            logger_1.default.warn(`WARN: Missing service implementations:\n\t${missing.join('\n\t')}`);
        }
    }
}
exports.GrpcAsyncServer = GrpcAsyncServer;
